#!/usr/bin/env python

# Question:  What should we do if all of the workers have a status
# of timedout?  Should we give up right away, or keep waiting for
# the "reply timeout"?

import sys, getopt, traceback, time, random
from nws.client import NwsServer
from nws.sleigh import Sleigh, sshcmd

R_TYPE = 0
A_TIME = 0
A_WORKER = 1
W_OP = 0
W_TASK = 1

class TaskTimeoutException(Exception):
    def __init__(self, task, worker):
        Exception.__init__(self)
        self._task = task
        self._worker = worker

    def task(self):
        return self._task

    def worker(self):
        return self._worker

    def __repr__(self):
        return "<Task timeout: %d; worker: %s>" % \
            (self.task(), str(self.worker()))

class GlobalTaskTimeoutException(TaskTimeoutException):
    def __init__(self, task, worker):
        TaskTimeoutException.__init__(self, task, worker)

    def __repr__(self):
        return "<Global task timeout: %d; worker: %s>" % \
            (self.task(), str(self.worker()))

class Timeout(object):
    def __init__(self, upper, lower, defValue=1000000.0):
        self._dataPoints = 0
        self._fudge = 0.25
        self._upper = max(upper - lower, lower)
        self._lower = lower
        self._max = 0.0
        self._timeout = defValue

    def dataPoint(self, taskTime):
        self._dataPoints += 1
        denom = (self._dataPoints - 1) * self._fudge + 1.0
        factor = (self._upper / denom) + self._lower
        if taskTime > self._max:
            self._max = taskTime
        self._timeout = factor * self._max

    def value(self):
        return self._timeout

    def __repr__(self):
        return "Timeout[cur:%f,max:%f,n:%d]" % \
            (self._timeout, self._max, self._dataPoints)

    def __str__(self):
        return self.__repr__()

def workerFunc(wsname):
    ws = SleighNws.server.useWs(wsname)

    while True:
        task = ws.fetch('task')

        if random.randint(0, 9) == 9:
            # XXX don't trigger this condition right now
            # break
            pass

        ws.store('result', ('allocated', task, SleighRank))

        result = 2 * task
        taskTime = random.randint(8, 10)
        time.sleep(taskTime)

        if random.randint(0, 30) == 30:
            break

        ws.store('result', ('completed', task, SleighRank, taskTime, result))

if __name__ == '__main__':
    numtasks = 10
    nkw = {}
    skw = {}

    try:
        opts, args = getopt.getopt(sys.argv[1:], 'h:p:n:c:v')

        for opt, arg in opts:
            if opt == '-h':
                nkw['host'] = skw['nwsHost'] = arg
            elif opt == '-p':
                nkw['port'] = skw['nwsPort'] = int(arg)
            elif opt == '-n':
                numtasks = int(arg)
            elif opt == '-c':
                skw['workerCount'] = int(arg)
            elif opt == '-v':
                skw['verbose'] = 1
            else:
                raise Exception('internal error: out-of-sync with getopt')

        if not skw.get('launch'):
            if len(args) > 0:
                skw['launch'] = sshcmd
                skw['nodeList'] = args
                if skw.has_key('workerCount'):
                    print >> sys.stderr, 'warning: ignoring -c option ' \
                        'since workers are specified'
        else:
            if len(args) > 0:
                print >> sys.stderr, \
                    'warning: ignoring unused arguments:', " ".join(args)
            print >> sys.stderr, 'starting sleigh in web launch mode'

        # create a work space to communicate with the workers
        server = NwsServer(**nkw)
        wsname = server.mktempWs("Tolerant_%d")
        ws = server.openWs(wsname)

        s = Sleigh(**skw)

        # need to set the __module__ attribute, because otherwise
        # our copy is set to '__main__', which is useless
        workerFunc.__module__ = 'tolerant'

        s.eachWorker(workerFunc, wsname, blocking=False)

        for i in range(numtasks):
            ws.store('task', i)

        taskTimeout = Timeout(20.0, 4.0, 200.0)
        lastReply = time.time()
        nextToTimeout = 0
        nextToAllocate = 0
        allocated = {}
        workerStatus = {}
        results = [None] * numtasks
        numResults = 0

        while numResults < numtasks:
            result = ws.fetchTry('result')
            now = time.time()

            if result is None:
                # if you don't get any replies for a long time, it may
                # be that everyone has died (or been disconnected from
                # the nws server due to a network failure).
                if now >= lastReply + 2.0 * taskTimeout.value():
                    print >> sys.stderr, "giving up because everyone seems dead"
                    for i in xrange(nextToTimeout, numtasks):
                        if results[i] is None:
                            if allocated.has_key(i):
                                worker = allocated[i][A_WORKER]
                            else:
                                worker = None
                            results[i] = GlobalTaskTimeoutException(i, worker)
                    allocated.clear()
                    numResults = numtasks
                    break
                else:
                    time.sleep(3)
            else:
                lastReply = now
                if result[R_TYPE] == 'allocated':
                    # Mark this and any previous tasks as allocated
                    op, task, worker = result
                    print >> sys.stderr, "worker %d allocated task %d" % \
                        (worker, task)
                    if allocated.has_key(task) and \
                            allocated[task][A_WORKER] is not None:
                        print >> sys.stderr, \
                            "!!!! task %d was already allocated by worker %s" % \
                            (task, str(allocated[task][A_WORKER]))
                    if workerStatus.has_key(worker) and \
                            workerStatus[worker][W_OP] == 'allocated':
                        print >> sys.stderr, \
                            "!!!! worker %d still has task %d allocated" % \
                            (worker, workerStatus[worker][W_TASK])
                    if task >= nextToAllocate:
                        for i in range(nextToAllocate, task + 1):
                            if i == task:
                                print >> sys.stderr, \
                                    "marking task %d as allocated by worker %d" % \
                                    (i, worker)
                                allocated[i] = [now, worker]
                                workerStatus[worker] = (op, task)
                            else:
                                print >> sys.stderr, \
                                    "marking task %d as allocated by None" % i
                                allocated[i] = [now, None]
                        nextToAllocate = task + 1
                    else:
                        # update the allocated record to include the worker info
                        print >> sys.stderr, "**** got late allocation message"
                        assert allocated.has_key(task) and \
                            allocated[task][A_WORKER] is None
                        allocated[task][A_WORKER] = worker
                        workerStatus[worker] = (op, task)
                elif result[R_TYPE] == 'completed':
                    op, task, worker, taskTime, val = result
                    print >> sys.stderr, "worker %d completed task %d: " % \
                        (worker, task), val
                    if not allocated.has_key(task):
                        print >> sys.stderr, \
                            "!!!! got results for unallocated task"
                    elif allocated[task][A_WORKER] != worker:
                        print >> sys.stderr, \
                            "!!!! worker %d completed a task he didn't allocate" % \
                            (worker,)
                    if not workerStatus.has_key(worker):
                        print >> sys.stderr, \
                            "!!!! got results from unknown worker"
                    elif workerStatus[worker] != ('allocated', task):
                        print >> sys.stderr, \
                            "!!!! worker %d had unexpected status: %s" % \
                            (worker, workerStatus[worker])
                    taskTimeout.dataPoint(taskTime)
                    results[task] = val
                    numResults += 1
                    workerStatus[worker] = (op, task)
                    del allocated[task]
                else:
                    print >> sys.stderr, "error: illegal result type:", \
                        result[R_TYPE]

            if len(allocated) > len(workerStatus):
                # this can happen if a worker loses the allocation race
                # on his very first task.  after his first task, he's in
                # the workerStatus dictionary.
                print >> sys.stderr, "**** more tasks allocated than workers"

            print >> sys.stderr, "allocated tasks:", allocated
            print >> sys.stderr, "worker status:", workerStatus

            for i in xrange(nextToTimeout, nextToAllocate):
                if not allocated.has_key(i):
                    # this task was already returned
                    nextToTimeout = i + 1
                elif now >= allocated[i][A_TIME] + taskTimeout.value():
                    print >> sys.stderr, \
                        "timing out task %d allocated by worker %s" % \
                        (i, str(allocated[i][A_WORKER]))
                    results[i] = TaskTimeoutException(i, allocated[i][A_WORKER])
                    numResults += 1
                    worker = allocated[i][A_WORKER]
                    if worker is not None:
                        if workerStatus[worker] != ('allocated', i):
                            print >> sys.stderr, \
                                "!!!! worker %d had unexpected status: %s" % \
                                (worker, workerStatus[worker])
                        workerStatus[worker] = ('timedout', i)
                    else:
                        print >> sys.stderr, \
                            "timed out task had unknown allocation"
                    del allocated[i]
                    nextToTimeout = i + 1
                else:
                    # if this one hasn't timed out, none of the rest
                    # have either, so break out of the loop.
                    # this is true because tasks are stored into a fifo
                    # mode variable.
                    break

        print results
    except getopt.GetoptError, e:
        print >> sys.stderr, e.msg
        sys.exit(1)
    except ValueError, e:
        print >> sys.stderr, "option %s requires an integer argument" % opt
        sys.exit(1)
    except (KeyboardInterrupt, SystemExit):
        pass
    except:
        traceback.print_exc()
        ex = sys.exc_info()
        print >> sys.stderr, ex[0], ex[1]
