#
# Copyright (c) 2005-2006, Scientific Computing Associates, Inc.
#
# NetWorkSpaces is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
# USA
#

"""Enhanced Python API for parallel programming using NetWorkSpaces.
"""

import time
import nws.sleigh
from nws.client import NetWorkSpace, NwsServer

__all__ = [
    'Sleigh'
]

class Sleigh(nws.sleigh.Sleigh):
    def __init__(self, *args, **kw):
        nws.sleigh.Sleigh.__init__(self, *args, **kw)

    def status(self, closeGroup=False, timeout=0.0):
        """Return the status of the worker group.

        s.status(closeGroup, timeout) -> numworkers, closed

        The status includes the number of workers that have joined the
        group so far, and a flag that indicates whether the group has
        been closed (meaning that no more workers can join).  Normally,
        the group is automatically closed when all the workers that were
        listed in the constructor have joined.  However, this method
        allows you to force the group to close after the timeout
        expires.  This can be particularly useful if you are running
        on a large number of nodes, and some of the nodes are slow or
        unreliable.  If some of the workers are never started, the group
        will never close, and no tasks will ever execute.

        Arguments:

        closeGroup -- Boolean flag indicating whether to close the
                group.  If True, the group will be closed, after the
                specified timeout.  The default value is False.

        timeout -- Number of seconds to wait for the group to close
                before returning.  The default value is 0.0.

        """
        if self.rankCount < 0:
            # join phase completed before
            return self.workerCount, 1
        else:
            sleepTime = initialSleep = min(2.0, timeout)
            repeatedSleep = 1.0
            startTime = time.time()

            while 1:
                n = self.nws.fetch('rankCount')
                if n < 0:
                    # all workers joined
                    if self.options['verbose']:
                        print 'all %d workers have started' % self.workerCount

                    assert self.workerCount == self.nws.find('workerCount')
                    self.rankCount = -1
                    self.nws.store('rankCount', self.rankCount)
                    return self.workerCount, 1
                else:
                    # three choices: close now, return not closed, or
                    # sleep and try again
                    if time.time() - startTime >= timeout:
                        # time is up, so either close the join, or
                        # return the current status
                        if closeGroup:
                            if self.options['verbose']:
                                print 'closing group: %d workers' % n

                            self.workerCount = n
                            self.nws.store('workerCount', self.workerCount)
                            self.rankCount = -1
                            self.nws.store('rankCount', self.rankCount)
                            return self.workerCount, 1
                        else:
                            if self.options['verbose']:
                                print 'group not formed: %d workers' %n

                            self.rankCount = n
                            self.nws.store('rankCount', self.rankCount)
                            return self.rankCount, 0
                    else:
                        if self.options['verbose']:
                            print 'only %d workers: sleeping...' % n

                        self.rankCount = n
                        self.nws.store('rankCount', n)
                        time.sleep(sleepTime)
                        sleepTime = repeatedSleep

    def __str__(self):
        return "ESleigh [%s] on nodes: %s" % \
            (str(self.nws), " ".join(self.nodeList))
